@@ -2,9 +2,12 @@ import logging
 import time
 from urllib.parse import urlparse
 
-from cached_property import threaded_cached_property
 import dns.resolver
+from cached_property import threaded_cached_property
 
+from .cache import autodiscover_cache
+from .properties import Autodiscover
+from .protocol import AutodiscoverProtocol
 from ..configuration import Configuration
 from ..credentials import OAuth2Credentials
 from ..errors import AutoDiscoverFailed, AutoDiscoverCircularRedirect, TransportError, RedirectError, UnauthorizedError
@@ -13,9 +16,6 @@ from ..transport import get_auth_method_from_response, DEFAULT_HEADERS, NOAUTH,
 from ..util import post_ratelimited, get_domain, get_redirect_url, _back_off_if_needed, _may_retry_on_error, \
     DummyResponse, CONNECTION_ERRORS, TLS_ERRORS
 from ..version import Version
-from .cache import autodiscover_cache
-from .properties import Autodiscover
-from .protocol import AutodiscoverProtocol
 
 log = logging.getLogger(__name__)
 
@@ -27,7 +27,7 @@ def discover(email, credentials=None, auth_type=None, retry_policy=None):
 
 
 class SrvRecord:
-    """A container for autodiscover-related SRV records in DNS"""
+    """A container for autodiscover-related SRV records in DNS."""
 
     def __init__(self, priority, weight, port, srv):
         self.priority = priority
@@ -36,7 +36,7 @@ class SrvRecord:
         self.srv = srv
 
     def __eq__(self, other):
-        for k in self.__dict__.keys():
+        for k in self.__dict__:
             if getattr(self, k) != getattr(other, k):
                 return False
         return True
@@ -67,7 +67,6 @@ class Autodiscovery:
 
     WARNING: The autodiscover protocol is very complicated. If you have problems autodiscovering using this
     implementation, start by doing an official test at https://testconnectivity.microsoft.com
-
     """
 
     # When connecting to servers that may not be serving the correct endpoint, we should use a retry policy that does
@@ -75,17 +74,19 @@ class Autodiscovery:
     INITIAL_RETRY_POLICY = FailFast()
     RETRY_WAIT = 10  # Seconds to wait before retry on connection errors
     MAX_REDIRECTS = 10  # Maximum number of URL redirects before we give up
+    DNS_RESOLVER_KWARGS = {}
+    DNS_RESOLVER_ATTRS = {
+        'timeout': AutodiscoverProtocol.TIMEOUT,
+    }
 
     def __init__(self, email, credentials=None, auth_type=None, retry_policy=None):
         """
 
-        Args:
-          email: The email address to autodiscover
-          credentials: Credentials with authorization to make autodiscover lookups for this Account
+        :param email: The email address to autodiscover
+        :param credentials: Credentials with authorization to make autodiscover lookups for this Account
             (Default value = None)
-          auth_type:  (Default value = None)
-          retry_policy:  (Default value = None)
-
+        :param auth_type:  (Default value = None)
+        :param retry_policy:  (Default value = None)
         """
         self.email = email
         self.credentials = credentials
@@ -148,8 +149,9 @@ class Autodiscovery:
 
     @threaded_cached_property
     def resolver(self):
-        resolver = dns.resolver.Resolver()
-        resolver.timeout = AutodiscoverProtocol.TIMEOUT
+        resolver = dns.resolver.Resolver(**self.DNS_RESOLVER_KWARGS)
+        for k, v in self.DNS_RESOLVER_ATTRS.items():
+            setattr(resolver, k, v)
         return resolver
 
     def _build_response(self, ad_response):
@@ -210,9 +212,8 @@ class Autodiscovery:
         it does not redirect to non-SSL endpoints or SSL endpoints with invalid certificates, and that the redirect is
         not circular. Finally, we should fail after 10 redirects.
 
-        Args:
-          url:
-
+        :param url:
+        :return:
         """
         if url.lower() in self._urls_visited:
             log.warning('We have already tried this URL: %s', url)
@@ -241,10 +242,9 @@ class Autodiscovery:
         """Get auth type by tasting headers from the server. Do POST requests be default. HEAD is too error prone, and
         some servers are set up to redirect to OWA on all requests except POST to the autodiscover endpoint.
 
-        Args:
-          url:
-          method:  (Default value = 'post')
-
+        :param url:
+        :param method:  (Default value = 'post')
+        :return:
         """
         # We are connecting to untrusted servers here, so take necessary precautions.
         hostname = urlparse(url).netloc
@@ -263,7 +263,7 @@ class Autodiscovery:
         while True:
             _back_off_if_needed(self.INITIAL_RETRY_POLICY.back_off_until)
             log.debug('Trying to get response from %s', url)
-            with AutodiscoverProtocol.raw_session() as s:
+            with AutodiscoverProtocol.raw_session(url) as s:
                 try:
                     r = getattr(s, method)(**kwargs)
                     r.close()  # Release memory
@@ -276,6 +276,8 @@ class Autodiscovery:
                     total_wait = time.monotonic() - t_start
                     if _may_retry_on_error(response=r, retry_policy=self.INITIAL_RETRY_POLICY, wait=total_wait):
                         log.debug("Connection error on URL %s (retry %s, error: %s). Cool down", url, retry, e)
+                        # Don't respect the 'Retry-After' header. We don't know if this is a useful endpoint, and we
+                        # want autodiscover to be reasonably fast.
                         self.INITIAL_RETRY_POLICY.back_off(self.RETRY_WAIT)
                         retry += 1
                         continue
@@ -286,21 +288,19 @@ class Autodiscovery:
         except UnauthorizedError:
             # Failed to guess the auth type
             auth_type = NOAUTH
-        if r.status_code in (301, 302):
-            if 'location' in r.headers:
-                # Make the redirect URL absolute
-                try:
-                    r.headers['location'] = get_redirect_url(r)
-                except TransportError:
-                    del r.headers['location']
+        if r.status_code in (301, 302) and 'location' in r.headers:
+            # Make the redirect URL absolute
+            try:
+                r.headers['location'] = get_redirect_url(r)
+            except TransportError:
+                del r.headers['location']
         return auth_type, r
 
     def _get_authenticated_response(self, protocol):
         """Get a response by using the credentials provided. We guess the auth type along the way.
 
-        Args:
-          protocol:
-
+        :param protocol:
+        :return:
         """
         # Redo the request with the correct auth
         data = Autodiscover.payload(email=self.email)
@@ -322,11 +322,10 @@ class Autodiscovery:
         return r
 
     def _attempt_response(self, url):
-        """Returns a (is_valid_response, response) tuple
-
-        Args:
-          url:
+        """Return an (is_valid_response, response) tuple.
 
+        :param url:
+        :return:
         """
         self._urls_visited.append(url.lower())
         log.debug('Attempting to get a valid response from %s', url)
@@ -387,9 +386,8 @@ class Autodiscovery:
 
         The first three numbers in the service line are: priority, weight, port
 
-        Args:
-          hostname:
-
+        :param hostname:
+        :return:
         """
         log.debug('Attempting to get SRV records for %s', hostname)
         records = []
@@ -411,43 +409,39 @@ class Autodiscovery:
         return records
 
     def _step_1(self, hostname):
-        """The client sends an Autodiscover request to https://example.com/autodiscover/autodiscover.xml and then does
-        one of the following:
+        """Perform step 1, where the client sends an Autodiscover request to
+        https://example.com/autodiscover/autodiscover.xml and then does one of the following:
             * If the Autodiscover attempt succeeds, the client proceeds to step 5.
             * If the Autodiscover attempt fails, the client proceeds to step 2.
 
-        Args:
-          hostname:
-
+        :param hostname:
+        :return:
         """
         url = 'https://%s/Autodiscover/Autodiscover.xml' % hostname
         log.info('Step 1: Trying autodiscover on %r with email %r', url, self.email)
         is_valid_response, ad = self._attempt_response(url=url)
         if is_valid_response:
             return self._step_5(ad=ad)
-        else:
-            return self._step_2(hostname=hostname)
+        return self._step_2(hostname=hostname)
 
     def _step_2(self, hostname):
-        """The client sends an Autodiscover request to https://autodiscover.example.com/autodiscover/autodiscover.xml
-        and then does one of the following:
+        """Perform step 2, where the client sends an Autodiscover request to
+        https://autodiscover.example.com/autodiscover/autodiscover.xml and then does one of the following:
             * If the Autodiscover attempt succeeds, the client proceeds to step 5.
             * If the Autodiscover attempt fails, the client proceeds to step 3.
 
-        Args:
-          hostname:
-
+        :param hostname:
+        :return:
         """
         url = 'https://autodiscover.%s/Autodiscover/Autodiscover.xml' % hostname
         log.info('Step 2: Trying autodiscover on %r with email %r', url, self.email)
         is_valid_response, ad = self._attempt_response(url=url)
         if is_valid_response:
             return self._step_5(ad=ad)
-        else:
-            return self._step_3(hostname=hostname)
+        return self._step_3(hostname=hostname)
 
     def _step_3(self, hostname):
-        """The client sends an unauth'ed GET method request to
+        """Perform step 3, where the client sends an unauth'ed GET method request to
         http://autodiscover.example.com/autodiscover/autodiscover.xml (Note that this is a non-HTTPS endpoint). The
         client then does one of the following:
             * If the GET request returns a 302 redirect response, it gets the redirection URL from the 'Location' HTTP
@@ -459,9 +453,8 @@ class Autodiscovery:
                 * If the redirection URL is not valid, the client proceeds to step 4.
             * If the GET request does not return a 302 redirect response, the client proceeds to step 4.
 
-        Args:
-          hostname:
-
+        :param hostname:
+        :return:
         """
         url = 'http://autodiscover.%s/Autodiscover/Autodiscover.xml' % hostname
         log.info('Step 3: Trying autodiscover on %r with email %r', url, self.email)
@@ -475,17 +468,15 @@ class Autodiscovery:
                 is_valid_response, ad = self._attempt_response(url=redirect_url)
                 if is_valid_response:
                     return self._step_5(ad=ad)
-                else:
-                    return self._step_4(hostname=hostname)
-            else:
                 return self._step_4(hostname=hostname)
-        else:
             return self._step_4(hostname=hostname)
+        return self._step_4(hostname=hostname)
 
     def _step_4(self, hostname):
-        """The client performs a Domain Name System (DNS) query for an SRV record for _autodiscover._tcp.example.com.
-        The query might return multiple records. The client selects only records that point to an SSL endpoint and that
-        have the highest priority and weight. One of the following actions then occurs:
+        """Perform step 4, where the client performs a Domain Name System (DNS) query for an SRV record for
+        _autodiscover._tcp.example.com. The query might return multiple records. The client selects only records that
+        point to an SSL endpoint and that have the highest priority and weight. One of the following actions then
+        occurs:
             * If no such records are returned, the client proceeds to step 6.
             * If records are returned, the application randomly chooses a record in the list and validates the endpoint
               that it points to by following the process described in the "Redirect Response" section. The client then
@@ -495,9 +486,8 @@ class Autodiscovery:
                     * If the attempt fails, the client proceeds to step 6.
                 * If the redirection URL is not valid, the client proceeds to step 6.
 
-        Args:
-          hostname:
-
+        :param hostname:
+        :return:
         """
         dns_hostname = '_autodiscover._tcp.%s' % hostname
         log.info('Step 4: Trying autodiscover on %r with email %r', dns_hostname, self.email)
@@ -508,19 +498,16 @@ class Autodiscovery:
             srv_host = None
         if not srv_host:
             return self._step_6()
-        else:
-            redirect_url = 'https://%s/Autodiscover/Autodiscover.xml' % srv_host
-            if self._redirect_url_is_valid(url=redirect_url):
-                is_valid_response, ad = self._attempt_response(url=redirect_url)
-                if is_valid_response:
-                    return self._step_5(ad=ad)
-                else:
-                    return self._step_6()
-            else:
-                return self._step_6()
+        redirect_url = 'https://%s/Autodiscover/Autodiscover.xml' % srv_host
+        if self._redirect_url_is_valid(url=redirect_url):
+            is_valid_response, ad = self._attempt_response(url=redirect_url)
+            if is_valid_response:
+                return self._step_5(ad=ad)
+            return self._step_6()
+        return self._step_6()
 
     def _step_5(self, ad):
-        """When a valid Autodiscover request succeeds, the following sequence occurs:
+        """Perform step 5. When a valid Autodiscover request succeeds, the following sequence occurs:
             * If the server responds with an HTTP 302 redirect, the client validates the redirection URL according to
               the process defined in the "Redirect responses" and then does one of the following:
                 * If the redirection URL is valid, the client tries the URL and then does one of the following:
@@ -533,9 +520,8 @@ class Autodiscovery:
                 * If the value of the Action element is "Settings", the client has successfully received the requested
                   configuration settings for the specified user. The client does not need to proceed to step 6.
 
-        Args:
-          ad:
-
+        :param ad:
+        :return:
         """
         log.info('Step 5: Checking response')
         if ad.response is None:
@@ -551,19 +537,16 @@ class Autodiscovery:
                 is_valid_response, ad = self._attempt_response(url=ad_response.redirect_url)
                 if is_valid_response:
                     return self._step_5(ad=ad)
-                else:
-                    return self._step_6()
-            else:
-                log.debug('Invalid redirect URL: %s', ad_response.redirect_url)
                 return self._step_6()
-        else:
-            # This could be an email redirect. Let outer layer handle this
-            return ad_response
+            log.debug('Invalid redirect URL: %s', ad_response.redirect_url)
+            return self._step_6()
+        # This could be an email redirect. Let outer layer handle this
+        return ad_response
 
     def _step_6(self):
-        """If the client cannot contact the Autodiscover service, the client should ask the user for the Exchange server
-        name and use it to construct an Exchange EWS URL. The client should try to use this URL for future requests.
-
+        """Perform step 6. If the client cannot contact the Autodiscover service, the client should ask the user for
+        the Exchange server name and use it to construct an Exchange EWS URL. The client should try to use this URL for
+        future requests.
         """
         raise AutoDiscoverFailed(
             'All steps in the autodiscover protocol failed for email %r. If you think this is an error, consider doing '
@@ -571,11 +554,10 @@ class Autodiscovery:
 
 
 def _select_srv_host(srv_records):
-    """Select the record with the highest priority, that also supports TLS
-
-    Args:
-      srv_records:
+    """Select the record with the highest priority, that also supports TLS.
 
+    :param srv_records:
+    :return:
     """
     best_record = None
     for srv_record in srv_records:
