@@ -1,7 +1,5 @@
 # flake8: noqa
-"""
-Stores errors specific to this package, and mirrors all the possible errors that EWS can return.
-"""
+"""Stores errors specific to this package, and mirrors all the possible errors that EWS can return."""
 from urllib.parse import urlparse
 
 
@@ -98,7 +96,12 @@ class AutoDiscoverRedirect(AutoDiscoverError):
 
 
 class NaiveDateTimeNotAllowed(ValueError):
-    pass
+    def __init__(self, local_dt):
+        super().__init__()
+        from .ewsdatetime import EWSDateTime
+        if not isinstance(local_dt, EWSDateTime):
+            raise ValueError("'local_dt' value %r must be an EWSDateTime" % local_dt)
+        self.local_dt = local_dt
 
 
 class UnknownTimeZone(EWSError):
@@ -120,7 +123,6 @@ class ResponseMessageError(TransportError):
 class CASError(EWSError):
     """EWS will sometimes return an error message in an 'X-CasErrorCode' custom HTTP header in an HTTP 500 error code.
     This exception is for those cases. The caller may want to do something with the original response, so store that.
-
     """
 
     def __init__(self, cas_error, response):
