@@ -4,11 +4,12 @@ A protocol is an endpoint for EWS service connections. It contains all necessary
 Protocols should be accessed through an Account, and are either created from a default Configuration or autodiscovered
 when creating an Account.
 """
+import abc
 import datetime
 import logging
 import os
-from threading import Lock
 from queue import LifoQueue, Empty, Full
+from threading import Lock
 
 import requests.adapters
 import requests.sessions
@@ -18,7 +19,7 @@ from requests_oauthlib import OAuth2Session
 
 from .credentials import OAuth2AuthorizationCodeCredentials, OAuth2Credentials
 from .errors import TransportError, SessionPoolMinSizeReached, SessionPoolMaxSizeReached
-from .properties import FreeBusyViewOptions, MailboxData, TimeWindow, TimeZone
+from .properties import FreeBusyViewOptions, MailboxData, TimeWindow, TimeZone, RoomList, DLMailbox
 from .services import GetServerTimeZones, GetRoomLists, GetRooms, ResolveNames, GetUserAvailability, \
     GetSearchableMailboxes, ExpandDL, ConvertId
 from .transport import get_auth_instance, get_service_authtype, NTLM, OAUTH2, CREDENTIALS_REQUIRED, DEFAULT_HEADERS
@@ -32,7 +33,7 @@ def close_connections():
 
 
 class BaseProtocol:
-    """Base class for Protocol which implements the bare essentials"""
+    """Base class for Protocol which implements the bare essentials."""
 
     # The maximum number of sessions (== TCP connections, see below) we will open to this service endpoint. Keep this
     # low unless you have an agreement with the Exchange admin on the receiving end to hammer the server and
@@ -46,7 +47,7 @@ class BaseProtocol:
     CONNECTIONS_PER_SESSION = 1
     # The number of times a session may be reused before creating a new session object. 'None' means "infinite".
     # Discarding sessions after a certain number of usages may limit memory leaks in the Session object.
-    MAX_SESSION_USAGE_COUNT = 100
+    MAX_SESSION_USAGE_COUNT = None
     # Timeout for HTTP requests
     TIMEOUT = 120
 
@@ -63,16 +64,13 @@ class BaseProtocol:
         if not config.service_endpoint:
             raise AttributeError("'config.service_endpoint' must be set")
         self.config = config
-        self._session_pool_size = self.SESSION_POOLSIZE
-
-        # Autodetect authentication type if necessary
-        if self.config.auth_type is None:
-            self.config.auth_type = self.get_auth_type()
+        self._session_pool_size = 0
+        self._session_pool_maxsize = config.max_connections or self.SESSION_POOLSIZE
 
         # Try to behave nicely with the remote server. We want to keep the connection open between requests.
         # We also want to re-use sessions, to avoid the NTLM auth handshake on every request. We must know the
-        # authentication method to create a session pool.
-        self._session_pool = self._create_session_pool()
+        # authentication method to create sessions.
+        self._session_pool = LifoQueue()
         self._session_pool_lock = Lock()
 
     @property
@@ -90,11 +88,10 @@ class BaseProtocol:
     @credentials.setter
     def credentials(self, value):
         # We are updating credentials, but that doesn't automatically propagate to the session objects. The simplest
-        # solution is to just kill the session pool and rebuild it.
+        # solution is to just kill the sessions in the pool.
         with self._session_pool_lock:
             self.config._credentials = value
             self.close()
-            self._session_pool = self._create_session_pool()
 
     @property
     def retry_policy(self):
@@ -114,7 +111,7 @@ class BaseProtocol:
     def __setstate__(self, state):
         # Restore the session pool and lock
         self.__dict__.update(state)
-        self._session_pool = self._create_session_pool()
+        self._session_pool = LifoQueue()
         self._session_pool_lock = Lock()
 
     def __del__(self):
@@ -129,7 +126,9 @@ class BaseProtocol:
         log.debug('Server %s: Closing sessions', self.server)
         while True:
             try:
-                self._session_pool.get(block=False).close()
+                session = self._session_pool.get(block=False)
+                self.close_session(session)
+                self._session_pool_size -= 1
             except Empty:
                 break
 
@@ -143,41 +142,28 @@ class BaseProtocol:
             max_retries=0,
         )
 
-    def get_auth_type(self):
-        # Autodetect and return authentication type
-        raise NotImplementedError()
-
-    def _create_session_pool(self):
-        # Create a pool to reuse sessions containing connections to the server
-        session_pool = LifoQueue()
-        for _ in range(self._session_pool_size):
-            session_pool.put(self.create_session(), block=False)
-        return session_pool
-
     @property
     def session_pool_size(self):
         return self._session_pool_size
 
     def increase_poolsize(self):
-        """Increases the session pool size. We increase by one session per call.
-        """
+        """Increases the session pool size. We increase by one session per call."""
         # Create a single session and insert it into the pool. We need to protect this with a lock while we are changing
-        # the pool size variable, to avoid race conditions. We must not exceed the SESSION_POOLSIZE limit.
-        if self._session_pool_size == self.SESSION_POOLSIZE:
+        # the pool size variable, to avoid race conditions. We must not exceed the pool size limit.
+        if self._session_pool_size == self._session_pool_maxsize:
             raise SessionPoolMaxSizeReached('Session pool size cannot be increased further')
         with self._session_pool_lock:
-            if self._session_pool_size >= self.SESSION_POOLSIZE:
+            if self._session_pool_size >= self._session_pool_maxsize:
                 log.debug('Session pool size was increased in another thread')
                 return
-            log.warning('Increasing session pool size from %s to %s', self._session_pool_size,
-                        self._session_pool_size + 1)
+            log.debug('Server %s: Increasing session pool size from %s to %s', self.server, self._session_pool_size,
+                      self._session_pool_size + 1)
             self._session_pool.put(self.create_session(), block=False)
             self._session_pool_size += 1
 
     def decrease_poolsize(self):
         """Decreases the session pool size in response to error messages from the server requesting to rate-limit
         requests. We decrease by one session per call.
-
         """
         # Take a single session from the pool and discard it. We need to protect this with a lock while we are changing
         # the pool size variable, to avoid race conditions. We must keep at least one session in the pool.
@@ -187,23 +173,35 @@ class BaseProtocol:
             if self._session_pool_size <= 1:
                 log.debug('Session pool size was decreased in another thread')
                 return
-            log.warning('Decreasing session pool size from %s to %s', self._session_pool_size,
+            log.warning('Server %s: Decreasing session pool size from %s to %s', self.server, self._session_pool_size,
                         self._session_pool_size - 1)
-            self.get_session().close()
+            session = self.get_session()
+            self.close_session(session)
             self._session_pool_size -= 1
 
     def get_session(self):
+        # Try to get a session from the queue. If the queue is empty, try to add one more session to the queue. If the
+        # queue is already at its max, wait until a session becomes available.
         _timeout = 60  # Rate-limit messages about session starvation
-        while True:
+        try:
+            session = self._session_pool.get(block=False)
+            log.debug('Server %s: Got session immediately', self.server)
+        except Empty:
             try:
-                log.debug('Server %s: Waiting for session', self.server)
-                session = self._session_pool.get(timeout=_timeout)
-                log.debug('Server %s: Got session %s', self.server, session.session_id)
-                session.usage_count += 1
-                return session
-            except Empty:
-                # This is normal when we have many worker threads starving for available sessions
-                log.debug('Server %s: No sessions available for %s seconds', self.server, _timeout)
+                self.increase_poolsize()
+            except SessionPoolMaxSizeReached:
+                pass
+            while True:
+                try:
+                    log.debug('Server %s: Waiting for session', self.server)
+                    session = self._session_pool.get(timeout=_timeout)
+                    break
+                except Empty:
+                    # This is normal when we have many worker threads starving for available sessions
+                    log.debug('Server %s: No sessions available for %s seconds', self.server, _timeout)
+        log.debug('Server %s: Got session %s', self.server, session.session_id)
+        session.usage_count += 1
+        return session
 
     def release_session(self, session):
         # This should never fail, as we don't have more sessions than the queue contains
@@ -216,18 +214,21 @@ class BaseProtocol:
         except Full:
             log.debug('Server %s: Session pool was already full %s', self.server, session.session_id)
 
+    @staticmethod
+    def close_session(session):
+        session.close()
+        del session
+
     def retire_session(self, session):
         # The session is useless. Close it completely and place a fresh session in the pool
         log.debug('Server %s: Retiring session %s', self.server, session.session_id)
-        session.close()
-        del session
+        self.close_session(session)
         self.release_session(self.create_session())
 
     def renew_session(self, session):
         # The session is useless. Close it completely and place a fresh session in the pool
         log.debug('Server %s: Renewing session %s', self.server, session.session_id)
-        session.close()
-        del session
+        self.close_session(session)
         return self.create_session()
 
     def refresh_credentials(self, session):
@@ -241,7 +242,7 @@ class BaseProtocol:
                 # they're the same as the session was created with. If
                 # this isn't the case, we can just go ahead with a new
                 # session using the already-updated credentials.
-                self.credentials.refresh()
+                self.credentials.refresh(session=session)
         return self.renew_session(session)
 
     def create_session(self):
@@ -250,7 +251,7 @@ class BaseProtocol:
         if self.credentials is None:
             if self.auth_type in CREDENTIALS_REQUIRED:
                 raise ValueError('Auth type %r requires credentials' % self.auth_type)
-            session = self.raw_session()
+            session = self.raw_session(self.service_endpoint)
             session.auth = get_auth_instance(auth_type=self.auth_type)
         else:
             with self.credentials.lock:
@@ -267,7 +268,7 @@ class BaseProtocol:
                         username = '\\' + self.credentials.username
                     else:
                         username = self.credentials.username
-                    session = self.raw_session()
+                    session = self.raw_session(self.service_endpoint)
                     session.auth = get_auth_instance(auth_type=self.auth_type, username=username,
                                                      password=self.credentials.password)
 
@@ -327,7 +328,7 @@ class BaseProtocol:
             token_url = 'https://login.microsoftonline.com/%s/oauth2/v2.0/token' % self.credentials.tenant_id
             client = BackendApplicationClient(client_id=self.credentials.client_id)
 
-        session = self.raw_session(oauth2_client=client, oauth2_session_params=session_params)
+        session = self.raw_session(self.service_endpoint, oauth2_client=client, oauth2_session_params=session_params)
         if not has_token:
             # Fetch the token explicitly -- it doesn't occur implicitly
             token = session.fetch_token(token_url=token_url, client_id=self.credentials.client_id,
@@ -341,15 +342,14 @@ class BaseProtocol:
         return session
 
     @classmethod
-    def raw_session(cls, oauth2_client=None, oauth2_session_params=None):
+    def raw_session(cls, prefix, oauth2_client=None, oauth2_session_params=None):
         if oauth2_client:
             session = OAuth2Session(client=oauth2_client, **(oauth2_session_params or {}))
         else:
             session = requests.sessions.Session()
         session.headers.update(DEFAULT_HEADERS)
         session.headers['User-Agent'] = cls.USERAGENT
-        session.mount('http://', adapter=cls.get_adapter())
-        session.mount('https://', adapter=cls.get_adapter())
+        session.mount(prefix, adapter=cls.get_adapter())
         return session
 
     def __repr__(self):
@@ -357,6 +357,8 @@ class BaseProtocol:
 
 
 class CachingProtocol(type):
+    """A metaclass for Protocol that caches Protocol instances based on a server+username key."""
+
     _protocol_cache = {}
     _protocol_cache_lock = Lock()
 
@@ -403,20 +405,30 @@ class CachingProtocol(type):
 
     @classmethod
     def clear_cache(mcs):
-        for key, protocol in mcs._protocol_cache.items():
-            if isinstance(protocol, Exception):
-                continue
-            service_endpoint = key[0]
-            log.debug("Service endpoint '%s': Closing sessions", service_endpoint)
-            protocol.close()
-        mcs._protocol_cache.clear()
+        with mcs._protocol_cache_lock:
+            for key, protocol in mcs._protocol_cache.items():
+                if isinstance(protocol, Exception):
+                    continue
+                service_endpoint = key[0]
+                log.debug("Service endpoint '%s': Closing sessions", service_endpoint)
+                with protocol._session_pool_lock:
+                    protocol.close()
+            mcs._protocol_cache.clear()
 
 
 class Protocol(BaseProtocol, metaclass=CachingProtocol):
+    """A class to handle all the low-level communication with an Exchange server. Contains a session pool, knows how to
+    negotiate the authentication type of the server, refresh credentials, etc. Also contains methods for calling EWS
+    services that are not tied to an account.
+    """
+
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self._api_version_hint = None
         self._version_lock = Lock()
+        # Autodetect authentication type if necessary
+        if self.config.auth_type is None:
+            self.config.auth_type = self.get_auth_type()
 
     def get_auth_type(self):
         # Autodetect authentication type. We also set version hint here.
@@ -438,35 +450,31 @@ class Protocol(BaseProtocol, metaclass=CachingProtocol):
         return self.config.version
 
     def get_timezones(self, timezones=None, return_full_timezone_data=False):
-        """Get timezone definitions from the server
+        """Get timezone definitions from the server.
 
-        Args:
-          timezones: A list of EWSDateTime instances. If None, fetches all timezones from server (Default value = None)
-          return_full_timezone_data: If true, also returns periods and transitions (Default value = False)
-
-        Returns:
-          A list of (tz_id, name, periods, transitions) tuples
+        :param timezones: A list of EWSDateTime instances. If None, fetches all timezones from server
+          (Default value = None)
+        :param return_full_timezone_data: If true, also returns periods and transitions (Default value = False)
 
+        :return: A list of (tz_id, name, periods, transitions) tuples
         """
         return GetServerTimeZones(protocol=self).call(
             timezones=timezones, return_full_timezone_data=return_full_timezone_data
         )
 
     def get_free_busy_info(self, accounts, start, end, merged_free_busy_interval=30, requested_view='DetailedMerged'):
-        """Returns free/busy information for a list of accounts
+        """Return free/busy information for a list of accounts.
 
-        Args:
-          accounts: A list of (account, attendee_type, exclude_conflicts) tuples, where account is either an Account
-            object or a string, attendee_type is a MailboxData.attendee_type choice, and exclude_conflicts is a boolean.
-          start: The start datetime of the request
-          end: The end datetime of the request
-          merged_free_busy_interval: The interval, in minutes, of merged free/busy information (Default value = 30)
-          requested_view: The type of information returned. Possible values are defined in the
-            FreeBusyViewOptions.requested_view choices. (Default value = 'DetailedMerged')
-
-        Returns:
-          A generator of FreeBusyView objects
+        :param accounts: A list of (account, attendee_type, exclude_conflicts) tuples, where account is either an
+          Account object or a string, attendee_type is a MailboxData.attendee_type choice, and exclude_conflicts is a
+          boolean.
+        :param start: The start datetime of the request
+        :param end: The end datetime of the request
+        :param merged_free_busy_interval: The interval, in minutes, of merged free/busy information (Default value = 30)
+        :param requested_view: The type of information returned. Possible values are defined in the
+          FreeBusyViewOptions.requested_view choices. (Default value = 'DetailedMerged')
 
+        :return: A generator of FreeBusyView objects
         """
         from .account import Account
         for account, attendee_type, exclude_conflicts in accounts:
@@ -510,22 +518,17 @@ class Protocol(BaseProtocol, metaclass=CachingProtocol):
         return GetRoomLists(protocol=self).call()
 
     def get_rooms(self, roomlist):
-        from .properties import RoomList
         return GetRooms(protocol=self).call(roomlist=RoomList(email_address=roomlist))
 
     def resolve_names(self, names, return_full_contact_data=False, search_scope=None, shape=None):
-        """Resolve accounts on the server using partial account data, e.g. an email address or initials
+        """Resolve accounts on the server using partial account data, e.g. an email address or initials.
 
-        Args:
-          names: A list of identifiers to query
-          return_full_contact_data: If True, returns full contact data (Default value = False)
-          search_scope: The scope to perform the search. Must be one of SEARCH_SCOPE_CHOICES (Default value = None)
-          shape: return: A list of Mailbox items or, if return_full_contact_data is True, tuples of (Mailbox, Contact)
-            items (Default value = None)
-
-        Returns:
-          A list of Mailbox items or, if return_full_contact_data is True, tuples of (Mailbox, Contact) items
+        :param names: A list of identifiers to query
+        :param return_full_contact_data: If True, returns full contact data (Default value = False)
+        :param search_scope: The scope to perform the search. Must be one of SEARCH_SCOPE_CHOICES (Default value = None)
+        :param shape: (Default value = None)
 
+        :return: A list of Mailbox items or, if return_full_contact_data is True, tuples of (Mailbox, Contact) items
         """
         return list(ResolveNames(protocol=self).call(
             unresolved_entries=names, return_full_contact_data=return_full_contact_data, search_scope=search_scope,
@@ -533,31 +536,26 @@ class Protocol(BaseProtocol, metaclass=CachingProtocol):
         ))
 
     def expand_dl(self, distribution_list):
-        """Expand distribution list into it's members
+        """Expand distribution list into it's members.
 
-        Args:
-          distribution_list: SMTP address of the distribution list to expand, or a DLMailbox representing the list
-
-        Returns:
-          List of Mailbox items that are members of the distribution list
+        :param distribution_list: SMTP address of the distribution list to expand, or a DLMailbox representing the list
 
+        :return: List of Mailbox items that are members of the distribution list
         """
-        from .properties import DLMailbox
         if isinstance(distribution_list, str):
             distribution_list = DLMailbox(email_address=distribution_list, mailbox_type='PublicDL')
         return list(ExpandDL(protocol=self).call(distribution_list=distribution_list))
 
     def get_searchable_mailboxes(self, search_filter=None, expand_group_membership=False):
-        """This method is only available to users who have been assigned the Discovery Management RBAC role. See
+        """Call the GetSearchableMailboxes service to get mailboxes that can be searched.
+
+        This method is only available to users who have been assigned the Discovery Management RBAC role. See
         https://docs.microsoft.com/en-us/exchange/permissions-exo/permissions-exo
 
-        Args:
-          search_filter: Is set, must be a single email alias (Default value = None)
-          expand_group_membership: If True, returned distribution lists are expanded (Default value = False)
-
-        Returns:
-          a list of SearchableMailbox, FailedMailbox or Exception instances
+        :param search_filter: Is set, must be a single email alias (Default value = None)
+        :param expand_group_membership: If True, returned distribution lists are expanded (Default value = False)
 
+        :return: a list of SearchableMailbox, FailedMailbox or Exception instances
         """
         return list(GetSearchableMailboxes(protocol=self).call(
             search_filter=search_filter,
@@ -565,26 +563,14 @@ class Protocol(BaseProtocol, metaclass=CachingProtocol):
         ))
 
     def convert_ids(self, ids, destination_format):
-        """Converts item and folder IDs between multiple formats
+        """Convert item and folder IDs between multiple formats.
 
-        Args:
-          ids: a list of AlternateId, AlternatePublicFolderId or AlternatePublicFolderItemId instances
-          destination_format: A string
-
-        Returns:
-          a generator of AlternateId, AlternatePublicFolderId or AlternatePublicFolderItemId instances
+        :param ids: a list of AlternateId, AlternatePublicFolderId or AlternatePublicFolderItemId instances
+        :param destination_format: A string
 
+        :return: a generator of AlternateId, AlternatePublicFolderId or AlternatePublicFolderItemId instances
         """
-        from .properties import AlternateId, AlternatePublicFolderId, AlternatePublicFolderItemId
-        cls_map = {cls.response_tag(): cls for cls in (
-            AlternateId, AlternatePublicFolderId, AlternatePublicFolderItemId
-        )}
-        for i in ConvertId(protocol=self).call(items=ids, destination_format=destination_format):
-            if isinstance(i, Exception):
-                yield i
-            else:
-                id_cls = cls_map[i.tag]
-                yield id_cls.from_xml(i, account=None)
+        return ConvertId(protocol=self).call(items=ids, destination_format=destination_format)
 
     def __getstate__(self):
         # The lock and thread pool cannot be pickled
@@ -622,12 +608,13 @@ class NoVerifyHTTPAdapter(requests.adapters.HTTPAdapter):
 
     def cert_verify(self, conn, url, verify, cert):
         # pylint: disable=unused-argument
-        # We're overiding a method so we have to keep the signature
+        # We're overriding a method so we have to keep the signature
         super().cert_verify(conn=conn, url=url, verify=False, cert=cert)
 
 
 class TLSClientAuth(requests.adapters.HTTPAdapter):
-    """An HTTP adapter that implements Certificate Based Authentication (CBA)"""
+    """An HTTP adapter that implements Certificate Based Authentication (CBA)."""
+
     cert_file = None
 
     def init_poolmanager(self, *args, **kwargs):
@@ -635,28 +622,34 @@ class TLSClientAuth(requests.adapters.HTTPAdapter):
         return super().init_poolmanager(*args, **kwargs)
 
 
-class RetryPolicy:
-    """Stores retry logic used when faced with errors from the server"""
+class RetryPolicy(metaclass=abc.ABCMeta):
+    """Stores retry logic used when faced with errors from the server."""
+
     @property
+    @abc.abstractmethod
     def fail_fast(self):
         # Used to choose the error handling policy. When True, a fault-tolerant policy is used. False, a fail-fast
         # policy is used.
-        raise NotImplementedError()
+        pass
 
     @property
+    @abc.abstractmethod
     def back_off_until(self):
-        raise NotImplementedError()
+        pass
 
     @back_off_until.setter
+    @abc.abstractmethod
     def back_off_until(self, value):
-        raise NotImplementedError()
+        pass
 
+    @abc.abstractmethod
     def back_off(self, seconds):
-        raise NotImplementedError()
+        pass
 
 
 class FailFast(RetryPolicy):
-    """Fail immediately on server errors"""
+    """Fail immediately on server errors."""
+
     @property
     def fail_fast(self):
         return True
@@ -665,11 +658,13 @@ class FailFast(RetryPolicy):
     def back_off_until(self):
         return None
 
+    def back_off(self, seconds):
+        raise ValueError('Cannot back off with fail-fast policy')
+
 
 class FaultTolerance(RetryPolicy):
     """Enables fault-tolerant error handling. Tells internal methods to do an exponential back off when requests start
     failing, and wait up to max_wait seconds before failing.
-
     """
 
     # Back off 60 seconds if we didn't get an explicit suggested value
@@ -697,7 +692,7 @@ class FaultTolerance(RetryPolicy):
 
     @property
     def back_off_until(self):
-        """Returns the back off value as a datetime. Resets the current back off value if it has expired."""
+        """Return the back off value as a datetime. Reset the current back off value if it has expired."""
         if self._back_off_until is None:
             return None
         with self._back_off_lock:
