@@ -1,21 +1,19 @@
-import logging
-
-from ..util import create_element, set_xml_value, MNS
 from .common import EWSAccountService, parse_folder_elem, to_item_id
-
-log = logging.getLogger(__name__)
+from ..fields import FieldPath
+from ..util import create_element, set_xml_value, MNS
 
 
 class UpdateFolder(EWSAccountService):
     """MSDN: https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/updatefolder-operation"""
+
     SERVICE_NAME = 'UpdateFolder'
     element_container_name = '{%s}Folders' % MNS
 
     def call(self, folders):
         # We can't easily find the correct folder class from the returned XML. Instead, return objects with the same
         # class as the folder instance it was requested with.
-        folders_list = list(f[0] for f in folders)  # Convert to a list, in case 'folders' is a generator
-        for folder, elem in zip(folders_list, self._get_elements(payload=self.get_payload(folders=folders))):
+        folders_list = list(folders)  # Convert to a list, in case 'folders' is a generator. We're iterating twice.
+        for (folder, _), elem in zip(folders_list, self._chunked_get_elements(self.get_payload, items=folders_list)):
             yield parse_folder_elem(elem=elem, folder=folder, account=self.account)
 
     @staticmethod
@@ -40,7 +38,6 @@ class UpdateFolder(EWSAccountService):
         return set_xml_value(deletefolderfield, field_path, version=self.account.version)
 
     def _get_folder_update_elems(self, folder, fieldnames):
-        from ..fields import FieldPath
         folder_model = folder.__class__
         fieldnames_set = set(fieldnames)
 
@@ -61,14 +58,13 @@ class UpdateFolder(EWSAccountService):
             yield self._set_folder_elem(folder_model=folder_model, field_path=FieldPath(field=field), value=value)
 
     def get_payload(self, folders):
-        from ..folders import BaseFolder, FolderId, DistinguishedFolderId
+        from ..folders import BaseFolder, FolderId
         updatefolder = create_element('m:%s' % self.SERVICE_NAME)
         folderchanges = create_element('m:FolderChanges')
         version = self.account.version
         for folder, fieldnames in folders:
-            log.debug('Updating folder %s fields %s', folder, fieldnames)
             folderchange = create_element('t:FolderChange')
-            if not isinstance(folder, (BaseFolder, FolderId, DistinguishedFolderId)):
+            if not isinstance(folder, (BaseFolder, FolderId)):
                 folder = to_item_id(folder, FolderId, version=version)
             set_xml_value(folderchange, folder, version=version)
             updates = create_element('t:Updates')
