@@ -7,7 +7,7 @@ import requests_oauthlib
 
 from .errors import UnauthorizedError, TransportError
 from .util import create_element, add_xml_child, xml_to_str, ns_translation, _may_retry_on_error, _back_off_if_needed, \
-    DummyResponse, CONNECTION_ERRORS
+    _retry_after, DummyResponse, CONNECTION_ERRORS
 
 log = logging.getLogger(__name__)
 
@@ -49,20 +49,6 @@ DEFAULT_ENCODING = 'utf-8'
 DEFAULT_HEADERS = {'Content-Type': 'text/xml; charset=%s' % DEFAULT_ENCODING, 'Accept-Encoding': 'gzip, deflate'}
 
 
-def extra_headers(primary_smtp_address):
-    """Generate extra HTTP headers
-
-    Args:
-      primary_smtp_address:
-
-    """
-    if primary_smtp_address:
-        # See
-        # https://blogs.msdn.microsoft.com/webdav_101/2015/05/11/best-practices-ews-authentication-and-access-issues/
-        return {'X-AnchorMailbox': primary_smtp_address}
-    return None
-
-
 def wrap(content, api_version, account_to_impersonate=None, timezone=None):
     """Generate the necessary boilerplate XML for a raw SOAP request. The XML is specific to the server version.
     ExchangeImpersonation allows to act as the user we want to impersonate.
@@ -76,12 +62,10 @@ def wrap(content, api_version, account_to_impersonate=None, timezone=None):
     TimeZoneContent element on MSDN:
     https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/timezonecontext
 
-    Args:
-      content:
-      api_version:
-      account_to_impersonate:  (Default value = None)
-      timezone:  (Default value = None)
-
+    :param content:
+    :param api_version:
+    :param account_to_impersonate:  (Default value = None)
+    :param timezone:  (Default value = None)
     """
     envelope = create_element('s:Envelope', nsmap=ns_translation)
     header = create_element('s:Header')
@@ -117,12 +101,10 @@ def wrap(content, api_version, account_to_impersonate=None, timezone=None):
 
 
 def get_auth_instance(auth_type, **kwargs):
-    """Returns an *Auth instance suitable for the requests package
-
-    Args:
-      auth_type:
-      **kwargs:
+    """Return an *Auth instance suitable for the requests package.
 
+    :param auth_type:
+    :param kwargs:
     """
     model = AUTH_TYPE_MAP[auth_type]
     if model is None:
@@ -153,7 +135,7 @@ def get_service_authtype(service_endpoint, retry_policy, api_versions, name):
         while True:
             _back_off_if_needed(retry_policy.back_off_until)
             log.debug('Trying to get service auth type for %s', service_endpoint)
-            with BaseProtocol.raw_session() as s:
+            with BaseProtocol.raw_session(service_endpoint) as s:
                 try:
                     r = s.post(url=service_endpoint, headers=headers, data=data, allow_redirects=False,
                                timeout=BaseProtocol.TIMEOUT)
@@ -164,13 +146,13 @@ def get_service_authtype(service_endpoint, retry_policy, api_versions, name):
                     total_wait = time.monotonic() - t_start
                     r = DummyResponse(url=service_endpoint, headers={}, request_headers=headers)
                     if _may_retry_on_error(response=r, retry_policy=retry_policy, wait=total_wait):
+                        wait = _retry_after(r, wait)
                         log.info("Connection error on URL %s (retry %s, error: %s). Cool down %s secs",
                                  service_endpoint, retry, e, wait)
                         retry_policy.back_off(wait)
                         retry += 1
                         continue
-                    else:
-                        raise TransportError(str(e)) from e
+                    raise TransportError(str(e)) from e
         if r.status_code not in (200, 401):
             log.debug('Unexpected response: %s %s', r.status_code, r.reason)
             continue
@@ -220,7 +202,7 @@ def _tokenize(val):
                 auth_methods.append(auth_method)
             auth_method = ''
             continue
-        elif c == '"':
+        if c == '"':
             auth_method += c
             if quote:
                 auth_methods.append(auth_method)
