@@ -1,33 +1,34 @@
-from base64 import b64decode
-from codecs import BOM_UTF8
-from collections import OrderedDict
 import datetime
-from decimal import Decimal
-from functools import wraps
 import io
 import itertools
 import logging
 import re
 import socket
-from threading import get_ident
 import time
-from urllib.parse import urlparse
 import xml.sax.handler  # nosec
+from base64 import b64decode, b64encode
+from codecs import BOM_UTF8
+from collections import OrderedDict
+from decimal import Decimal
+from functools import wraps
+from threading import get_ident
+from urllib.parse import urlparse
 
+import isodate
 import lxml.etree  # nosec
+import requests.exceptions
 from defusedxml.expatreader import DefusedExpatParser
 from defusedxml.sax import _InputSource
-import isodate
 from oauthlib.oauth2 import TokenExpiredError
 from pygments import highlight
-from pygments.lexers.html import XmlLexer
 from pygments.formatters.terminal import TerminalFormatter
-import requests.exceptions
+from pygments.lexers.html import XmlLexer
 
 from .errors import TransportError, RateLimitError, RedirectError, RelativeRedirect, CASError, UnauthorizedError, \
     ErrorInvalidSchemaVersionForMailboxVersion
 
 log = logging.getLogger(__name__)
+xml_log = logging.getLogger('%s.xml' % __name__)
 
 
 def require_account(f):
@@ -51,11 +52,12 @@ def require_id(f):
 
 
 class ParseError(lxml.etree.ParseError):
-    """Used to wrap lxml ParseError in our own class"""
-    pass
+    """Used to wrap lxml ParseError in our own class."""
 
 
 class ElementNotFound(Exception):
+    """Raised when the expected element was not found in a response stream."""
+
     def __init__(self, msg, data):
         super().__init__(msg)
         self.data = data
@@ -83,17 +85,14 @@ for item in ns_translation.items():
 
 
 def is_iterable(value, generators_allowed=False):
-    """Checks if value is a list-like object. Don't match generators and generator-like objects here by default, because
+    """Check if value is a list-like object. Don't match generators and generator-like objects here by default, because
     callers don't necessarily guarantee that they only iterate the value once. Take care to not match string types and
     bytes.
 
-    Args:
-      value: any type of object
-      generators_allowed: if True, generators will be treated as iterable (Default value = False)
-
-    Returns:
-      True or False
+    :param value: any type of object
+    :param generators_allowed: if True, generators will be treated as iterable (Default value = False)
 
+    :return: True or False
     """
     if generators_allowed:
         if not isinstance(value, (bytes, str)) and hasattr(value, '__iter__'):
@@ -105,12 +104,11 @@ def is_iterable(value, generators_allowed=False):
 
 
 def chunkify(iterable, chunksize):
-    """Splits an iterable into chunks of size ``chunksize``. The last chunk may be smaller than ``chunksize``.
-
-    Args:
-      iterable:
-      chunksize:
+    """Split an iterable into chunks of size ``chunksize``. The last chunk may be smaller than ``chunksize``.
 
+    :param iterable:
+    :param chunksize:
+    :return:
     """
     from .queryset import QuerySet
     if hasattr(iterable, '__getitem__') and not isinstance(iterable, QuerySet):
@@ -130,17 +128,11 @@ def chunkify(iterable, chunksize):
 
 
 def peek(iterable):
-    """Checks if an iterable is empty and returns status and the rewinded iterable
-
-    Args:
-      iterable:
+    """Check if an iterable is empty and return status and the rewinded iterable.
 
+    :param iterable:
+    :return:
     """
-    from .queryset import QuerySet
-    if isinstance(iterable, QuerySet):
-        # QuerySet has __len__ but that evaluates the entire query greedily. We don't want that here. Instead, peek()
-        # should be called on QuerySet.iterator()
-        raise ValueError('Cannot peek on a QuerySet')
     if hasattr(iterable, '__len__'):
         # tuple, list, set
         return not iterable, iterable
@@ -156,11 +148,10 @@ def peek(iterable):
 def xml_to_str(tree, encoding=None, xml_declaration=False):
     """Serialize an XML tree. Returns unicode if 'encoding' is None. Otherwise, we return encoded 'bytes'.
 
-    Args:
-      tree:
-      encoding:  (Default value = None)
-      xml_declaration:  (Default value = False)
-
+    :param tree:
+    :param encoding:  (Default value = None)
+    :param xml_declaration:  (Default value = False)
+    :return:
     """
     if xml_declaration and not encoding:
         raise ValueError("'xml_declaration' is not supported when 'encoding' is None")
@@ -181,14 +172,16 @@ def get_xml_attrs(tree, name):
 
 
 def value_to_xml_text(value):
-    # We can't handle bytes in this function because str == bytes on Python2
     from .ewsdatetime import EWSTimeZone, EWSDateTime, EWSDate
     from .indexed_properties import PhoneNumber, EmailAddress
-    from .properties import Mailbox, Attendee, ConversationId
+    from .properties import Mailbox, AssociatedCalendarItemId, Attendee, ConversationId
+    # We can't just create a map and look up with type(value) because we want to support subtypes
     if isinstance(value, str):
         return safe_xml_value(value)
     if isinstance(value, bool):
         return '1' if value else '0'
+    if isinstance(value, bytes):
+        return b64encode(value).decode('ascii')
     if isinstance(value, (int, Decimal)):
         return str(value)
     if isinstance(value, datetime.time):
@@ -209,19 +202,32 @@ def value_to_xml_text(value):
         return value.mailbox.email_address
     if isinstance(value, ConversationId):
         return value.id
-    raise NotImplementedError('Unsupported type: %s (%s)' % (type(value), value))
+    if isinstance(value, AssociatedCalendarItemId):
+        return value.id
+    raise TypeError('Unsupported type: %s (%s)' % (type(value), value))
 
 
 def xml_text_to_value(value, value_type):
-    # We can't handle bytes in this function because str == bytes on Python2
-    from .ewsdatetime import EWSDateTime
+    from .ewsdatetime import EWSDate, EWSDateTime
+    if value_type == str:
+        return value
+    if value_type == bool:
+        try:
+            return {
+                'true': True,
+                'on': True,
+                'false': False,
+                'off': False,
+            }[value.lower()]
+        except KeyError:
+            return None
     return {
-        bool: lambda v: True if v == 'true' else False if v == 'false' else None,
+        bytes: safe_b64decode,
         int: int,
         Decimal: Decimal,
         datetime.timedelta: isodate.parse_duration,
+        EWSDate: EWSDate.from_string,
         EWSDateTime: EWSDateTime.from_string,
-        str: lambda v: v
     }[value_type](value)
 
 
@@ -286,7 +292,6 @@ def add_xml_child(tree, name, value):
 class StreamingContentHandler(xml.sax.handler.ContentHandler):
     """A SAX content handler that returns a character data for a single element back to the parser. The parser must have
     a 'buffer' attribute we can append data to.
-
     """
 
     def __init__(self, parser, ns, element_name):
@@ -335,7 +340,7 @@ def safe_b64decode(data):
 
 
 class StreamingBase64Parser(DefusedExpatParser):
-    """A SAX parser that returns a generator of base64-decoded character content"""
+    """A SAX parser that returns a generator of base64-decoded character content."""
 
     def __init__(self, *args, **kwargs):
         DefusedExpatParser.__init__(self, *args, **kwargs)
@@ -356,8 +361,7 @@ class StreamingBase64Parser(DefusedExpatParser):
         while buffer:
             if not self.element_found:
                 collected_data += buffer
-            for data in self.feed(buffer):
-                yield data
+            yield from self.feed(buffer)
             buffer = file.read(self._bufsize)
         # Any remaining data in self.buffer should be padding chars now
         self.buffer = None
@@ -368,7 +372,7 @@ class StreamingBase64Parser(DefusedExpatParser):
             raise ElementNotFound('The element to be streamed from was not found', data=bytes(data))
 
     def feed(self, data, isFinal=0):
-        # Like upstream, but yields the current content of the character buffer
+        """Yield the current content of the character buffer."""
         DefusedExpatParser.feed(self, data=data, isFinal=isFinal)
         return self._decode_buffer()
 
@@ -399,7 +403,6 @@ class BytesGeneratorIO(io.RawIOBase):
     """A BytesIO that can produce bytes from a streaming HTTP request. Expects r.iter_content() as input
     lxml tries to be smart by calling `getvalue` when present, assuming that the entire string is in memory.
     Omitting `getvalue` forces lxml to stream the request through `read` avoiding the memory duplication.
-
     """
 
     def __init__(self, bytes_generator):
@@ -445,8 +448,57 @@ class BytesGeneratorIO(io.RawIOBase):
         super().close()
 
 
+class DocumentYielder:
+    """Look for XML documents in a streaming HTTP response and yield them as they become available from the stream."""
+
+    def __init__(self, content_iterator, document_tag='Envelope'):
+        self._iterator = content_iterator
+        self._start_token = b'<%s' % document_tag.encode('utf-8')
+        self._end_token = b'/%s>' % document_tag.encode('utf-8')
+
+    def get_tag(self, stop_byte):
+        tag_buffer = [b'<']
+        while True:
+            try:
+                c = next(self._iterator)
+            except StopIteration:
+                break
+            tag_buffer.append(c)
+            if c == stop_byte:
+                break
+        return b''.join(tag_buffer)
+
+    def __iter__(self):
+        """Consumes the content iterator, looking for start and end tags. Returns each document when we have fully
+        collected it.
+        """
+        doc_started = False
+        buffer = []
+        try:
+            while True:
+                c = next(self._iterator)
+                if not doc_started and c == b'<':
+                    tag = self.get_tag(stop_byte=b' ')
+                    if tag.startswith(self._start_token):
+                        # Start of document. Collect bytes from this point
+                        buffer.append(tag)
+                        doc_started = True
+                elif doc_started and c == b'<':
+                    tag = self.get_tag(stop_byte=b'>')
+                    buffer.append(tag)
+                    if tag.endswith(self._end_token):
+                        # End of document. Yield a valid document and reset the buffer
+                        yield b"<?xml version='1.0' encoding='utf-8'?>\n%s" % b''.join(buffer)
+                        doc_started = False
+                        buffer = []
+                elif doc_started:
+                    buffer.append(c)
+        except StopIteration:
+            return
+
+
 def to_xml(bytes_content):
-    # Converts bytes or a generator of bytes to an XML tree
+    """Convert bytes or a generator of bytes to an XML tree."""
     # Exchange servers may spit out the weirdest XML. lxml is pretty good at recovering from errors
     if isinstance(bytes_content, bytes):
         stream = io.BytesIO(bytes_content)
@@ -490,10 +542,9 @@ def to_xml(bytes_content):
 def is_xml(text, expected_prefix=b'<?xml'):
     """Lightweight test if response is an XML doc. It's better to be fast than correct here.
 
-    Args:
-      text: The string to check
-      expected_prefix: What to search for in the start if the string
-
+    :param text: The string to check
+    :param expected_prefix: What to search for in the start if the string
+    :return:
     """
     # BOM_UTF8 is an UTF-8 byte order mark which may precede the XML from an Exchange server
     bom_len = len(BOM_UTF8)
@@ -506,14 +557,15 @@ def is_xml(text, expected_prefix=b'<?xml'):
 
 
 class PrettyXmlHandler(logging.StreamHandler):
-    """A steaming log handler that prettifies log statements containing XML when output is a terminal"""
+    """A steaming log handler that prettifies log statements containing XML when output is a terminal."""
+
     @staticmethod
     def parse_bytes(xml_bytes):
         return lxml.etree.parse(io.BytesIO(xml_bytes), parser=_forgiving_parser)  # nosec
 
     @classmethod
     def prettify_xml(cls, xml_bytes):
-        # Re-formats an XML document to a consistent style
+        """Re-format an XML document to a consistent style."""
         return lxml.etree.tostring(
             cls.parse_bytes(xml_bytes),
             xml_declaration=True,
@@ -523,8 +575,8 @@ class PrettyXmlHandler(logging.StreamHandler):
 
     @staticmethod
     def highlight_xml(xml_str):
-        # Highlights a string containing XML, using terminal color codes
-        return highlight(xml_str, XmlLexer(), TerminalFormatter())
+        """Highlight a string containing XML, using terminal color codes."""
+        return highlight(xml_str, XmlLexer(), TerminalFormatter()).rstrip()
 
     def emit(self, record):
         """Pretty-print and syntax highlight a log statement if all these conditions are met:
@@ -532,9 +584,7 @@ class PrettyXmlHandler(logging.StreamHandler):
            * We're outputting to a terminal
            * The log message args is a dict containing keys starting with 'xml_' and values as bytes
 
-        Args:
-          record:
-
+        :param record:
         """
         if record.levelno == logging.DEBUG and self.is_tty() and isinstance(record.args, dict):
             for key, value in record.args.items():
@@ -552,7 +602,7 @@ class PrettyXmlHandler(logging.StreamHandler):
         return super().emit(record)
 
     def is_tty(self):
-        # Check if we're outputting to a terminal
+        """Check if we're outputting to a terminal."""
         try:
             return self.stream.isatty()
         except AttributeError:
@@ -560,7 +610,7 @@ class PrettyXmlHandler(logging.StreamHandler):
 
 
 class AnonymizingXmlHandler(PrettyXmlHandler):
-    """A steaming log handler that prettifies and anonymizes log statements containing XML when output is a terminal"""
+    """A steaming log handler that prettifies and anonymizes log statements containing XML when output is a terminal."""
 
     def __init__(self, forbidden_strings, *args, **kwargs):
         self.forbidden_strings = forbidden_strings
@@ -577,11 +627,15 @@ class AnonymizingXmlHandler(PrettyXmlHandler):
 
 
 class DummyRequest:
+    """A class to fake a requests Request object for functions that expect this."""
+
     def __init__(self, headers):
         self.headers = headers
 
 
 class DummyResponse:
+    """A class to fake a requests Response object for functions that expect this."""
+
     def __init__(self, url, headers, request_headers, content=b'', status_code=503, history=None):
         self.status_code = status_code
         self.url = url
@@ -659,7 +713,7 @@ except ImportError:
     pass
 
 
-def post_ratelimited(protocol, session, url, headers, data, allow_redirects=False, stream=False):
+def post_ratelimited(protocol, session, url, headers, data, allow_redirects=False, stream=False, timeout=None):
     """There are two error-handling policies implemented here: a fail-fast policy intended for stand-alone scripts which
     fails on all responses except HTTP 200. The other policy is intended for long-running tasks that need to respect
     rate-limiting errors from the server and paper over outages of up to 1 hour.
@@ -684,16 +738,19 @@ def post_ratelimited(protocol, session, url, headers, data, allow_redirects=Fals
     An additional resource on handling throttling policies and client back off strategies:
         https://docs.microsoft.com/en-us/exchange/client-developer/exchange-web-services/ews-throttling-in-exchange
 
-    Args:
-      protocol:
-      session:
-      url:
-      headers:
-      data:
-      allow_redirects:  (Default value = False)
-      stream:  (Default value = False)
+    :param protocol:
+    :param session:
+    :param url:
+    :param headers:
+    :param data:
+    :param allow_redirects:  (Default value = False)
+    :param stream:  (Default value = False)
+    :param timeout:
 
+    :return:
     """
+    if not timeout:
+        timeout = protocol.TIMEOUT
     thread_id = get_ident()
     wait = RETRY_WAIT  # Initial retry wait. We double the value on each retry
     retry = 0
@@ -712,14 +769,14 @@ Streaming: %(stream)s
 Response time: %(response_time)s
 Status code: %(status_code)s
 Request headers: %(request_headers)s
-Response headers: %(response_headers)s
-Request data: %(xml_request)s
-Response data: %(xml_response)s
-'''
+Response headers: %(response_headers)s'''
+    xml_log_msg = '''\
+Request XML: %(xml_request)s
+Response XML: %(xml_response)s'''
     log_vals = dict(
         retry=retry,
         wait=wait,
-        timeout=protocol.TIMEOUT,
+        timeout=timeout,
         session_id=session.session_id,
         thread_id=thread_id,
         auth=session.auth,
@@ -731,7 +788,9 @@ Response data: %(xml_response)s
         status_code=None,
         request_headers=headers,
         response_headers=None,
-        xml_request=data,
+    )
+    xml_log_vals = dict(
+        xml_request=None,
         xml_response=None,
     )
     t_start = time.monotonic()
@@ -742,12 +801,12 @@ Response data: %(xml_response)s
                 # We may have slept for a long time. Renew the session.
                 session = protocol.renew_session(session)
             log.debug('Session %s thread %s: retry %s timeout %s POST\'ing to %s after %ss wait', session.session_id,
-                      thread_id, retry, protocol.TIMEOUT, url, wait)
+                      thread_id, retry, timeout, url, wait)
             d_start = time.monotonic()
             # Always create a dummy response for logging purposes, in case we fail in the following
             r = DummyResponse(url=url, headers={}, request_headers=headers)
             try:
-                r = session.post(url=url, headers=headers, data=data, allow_redirects=False, timeout=protocol.TIMEOUT,
+                r = session.post(url=url, headers=headers, data=data, allow_redirects=False, timeout=timeout,
                                  stream=stream)
             except TLS_ERRORS as e:
                 # Don't retry on TLS errors. They will most likely be persistent.
@@ -773,9 +832,13 @@ Response data: %(xml_response)s
                     status_code=r.status_code,
                     request_headers=r.request.headers,
                     response_headers=r.headers,
+                )
+                xml_log_vals.update(
+                    xml_request=data,
                     xml_response='[STREAMING]' if stream else r.content,
                 )
             log.debug(log_msg, log_vals)
+            xml_log.debug(xml_log_msg, xml_log_vals)
             if _need_new_credentials(response=r):
                 r.close()  # Release memory
                 session = protocol.refresh_credentials(session)
@@ -785,6 +848,7 @@ Response data: %(xml_response)s
                 r.close()  # Release memory
                 log.info("Session %s thread %s: Connection error on URL %s (code %s). Cool down %s secs",
                          session.session_id, thread_id, r.url, r.status_code, wait)
+                wait = _retry_after(r, wait)
                 protocol.retry_policy.back_off(wait)
                 retry += 1
                 wait *= 2  # Increase delay for every retry
@@ -800,7 +864,7 @@ Response data: %(xml_response)s
         raise
     except Exception as e:
         # Let higher layers handle this. Add full context for better debugging.
-        log.error(str('%s: %s\n%s'), e.__class__.__name__, str(e), log_msg % log_vals)
+        log.error('%s: %s\n%s\n%s', e.__class__.__name__, str(e), log_msg % log_vals, xml_log_msg % xml_log_vals)
         protocol.retire_session(session)
         raise
     if r.status_code == 500 and r.content and is_xml(r.content):
@@ -808,7 +872,10 @@ Response data: %(xml_response)s
         log.debug('Got status code %s but trying to parse content anyway', r.status_code)
     elif r.status_code != 200:
         protocol.retire_session(session)
-        _raise_response_errors(r, protocol, log_msg, log_vals)  # Always raises an exception
+        try:
+            _raise_response_errors(r, protocol)  # Always raises an exception
+        finally:
+            log.error('%s\n%s', log_msg % log_vals, xml_log_msg % xml_log_vals)
     log.debug('Session %s thread %s: Useful response from %s', session.session_id, thread_id, url)
     return r, session
 
@@ -825,7 +892,7 @@ def _back_off_if_needed(back_off_until):
 
 
 def _may_retry_on_error(response, retry_policy, wait):
-    if response.status_code not in (301, 302, 401, 503):
+    if response.status_code not in (301, 302, 401, 500, 503):
         # Don't retry if we didn't get a status code that we can hope to recover from
         log.debug('No retry: wrong status code %s', response.status_code)
         return False
@@ -836,14 +903,24 @@ def _may_retry_on_error(response, retry_policy, wait):
         # We lost patience. Session is cleaned up in outer loop
         raise RateLimitError(
             'Max timeout reached', url=response.url, status_code=response.status_code, total_wait=wait)
-    # The genericerrorpage.htm/internalerror.asp is ridiculous behaviour for random outages. Redirect to
-    # '/internalsite/internalerror.asp' or '/internalsite/initparams.aspx' is caused by e.g. TLS certificate
-    # f*ckups on the Exchange server.
-    if (response.status_code == 401) \
-            or (response.headers.get('connection') == 'close') \
-            or (response.status_code == 302 and response.headers.get('location', '').lower() ==
-                '/ews/genericerrorpage.htm?aspxerrorpath=/ews/exchange.asmx') \
-            or (response.status_code == 503):
+    if response.status_code == 401:
+        # EWS sometimes throws 401's when it wants us to throttle connections. OK to retry.
+        return True
+    if response.headers.get('connection') == 'close':
+        # Connection closed. OK to retry.
+        return True
+    if response.status_code == 302 and response.headers.get('location', '').lower() \
+            == '/ews/genericerrorpage.htm?aspxerrorpath=/ews/exchange.asmx':
+        # The genericerrorpage.htm/internalerror.asp is ridiculous behaviour for random outages. OK to retry.
+        #
+        # Redirect to '/internalsite/internalerror.asp' or '/internalsite/initparams.aspx' is caused by e.g. TLS
+        # certificate f*ckups on the Exchange server. We should not retry those.
+        return True
+    if response.status_code == 503:
+        # Internal server error. OK to retry.
+        return True
+    if response.status_code == 500 and b"Server Error in '/EWS' Application" in response.content:
+        # "Server Error in '/EWS' Application" has been seen in highly concurrent settings. OK to retry.
         log.debug('Retry allowed: conditions met')
         return True
     return False
@@ -871,7 +948,7 @@ def _redirect_or_fail(response, redirects, allow_redirects):
     return redirect_url, redirects
 
 
-def _raise_response_errors(response, protocol, log_msg, log_vals):
+def _raise_response_errors(response, protocol):
     cas_error = response.headers.get('X-CasErrorCode')
     if cas_error:
         if cas_error.startswith('CAS error:'):
@@ -880,6 +957,7 @@ def _raise_response_errors(response, protocol, log_msg, log_vals):
         raise CASError(cas_error=cas_error, response=response)
     if response.status_code == 500 and (b'The specified server version is invalid' in response.content or
                                         b'ErrorInvalidSchemaVersionForMailboxVersion' in response.content):
+        # Another way of communicating invalid schema versions
         raise ErrorInvalidSchemaVersionForMailboxVersion('Invalid server version')
     if b'The referenced account is currently locked out' in response.content:
         raise TransportError('The service account is currently locked out')
@@ -887,6 +965,22 @@ def _raise_response_errors(response, protocol, log_msg, log_vals):
         # This is a login failure
         raise UnauthorizedError('Invalid credentials for %s' % response.url)
     if 'TimeoutException' in response.headers:
+        # A header set by us on CONNECTION_ERRORS
         raise response.headers['TimeoutException']
-    # This could be anything. Let higher layers handle this. Add full context for better debugging.
-    raise TransportError(str('Unknown failure\n') + log_msg % log_vals)
+    # This could be anything. Let higher layers handle this
+    raise TransportError(
+        'Unknown failure in response. Code: %s headers: %s content: %s'
+        % (response.status_code, response.headers, response.text)
+    )
+
+
+def _retry_after(r, wait):
+    """Either return the Retry-After header value or the default wait, whichever is larger."""
+    try:
+        retry_after = int(r.headers.get('Retry-After', '0'))
+    except ValueError:
+        pass
+    else:
+        if retry_after > wait:
+            return retry_after
+    return wait
