@@ -1,15 +1,9 @@
-# coding=utf-8
-from __future__ import unicode_literals
-
 import logging
 import re
 
-from future.utils import python_2_unicode_compatible
-from six import text_type, string_types
-
 from .errors import TransportError, ErrorInvalidSchemaVersionForMailboxVersion, ErrorInvalidServerVersion, \
     ErrorIncorrectSchemaVersion, ResponseMessageError
-from .util import PickleMixIn, xml_to_str, TNS
+from .util import xml_to_str, TNS
 
 log = logging.getLogger(__name__)
 
@@ -44,11 +38,8 @@ VERSIONS = {
 API_VERSIONS = sorted({v[0] for v in VERSIONS.values()}, reverse=True)
 
 
-@python_2_unicode_compatible
-class Build(PickleMixIn):
-    """
-    Holds methods for working with build numbers
-    """
+class Build:
+    """Holds methods for working with build numbers."""
 
     # List of build numbers here: https://docs.microsoft.com/en-us/exchange/new-features/build-numbers-and-release-dates
     API_VERSION_MAP = {
@@ -72,7 +63,7 @@ class Build(PickleMixIn):
         },
     }
 
-    __slots__ = ('major_version', 'minor_version', 'major_build', 'minor_build')
+    __slots__ = 'major_version', 'minor_version', 'major_build', 'minor_build'
 
     def __init__(self, major_version, minor_version, major_build=0, minor_build=0):
         if not isinstance(major_version, int):
@@ -88,7 +79,7 @@ class Build(PickleMixIn):
         self.major_build = major_build
         self.minor_build = minor_build
         if major_version < 8:
-            raise ValueError("Exchange major versions below 8 don't support EWS (%s)" % text_type(self))
+            raise ValueError("Exchange major versions below 8 don't support EWS (%s)" % self)
 
     @classmethod
     def from_xml(cls, elem):
@@ -106,6 +97,26 @@ class Build(PickleMixIn):
             kwargs[k] = int(v)  # Also raises ValueError
         return cls(**kwargs)
 
+    @classmethod
+    def from_hex_string(cls, s):
+        """Parse a server version string as returned in an autodiscover response. The process is described here:
+        https://docs.microsoft.com/en-us/exchange/client-developer/web-service-reference/serverversion-pox#example
+
+        The string is a hex string that, converted to a 32-bit binary, encodes the server version. The rules are:
+            * The first 4 bits contain the version number structure version. Can be ignored
+            * The next 6 bits contain the major version number
+            * The next 6 bits contain the minor version number
+            * The next bit contains a flag. Can be ignored
+            * The next 15 bits contain the major build number
+
+        :param s:
+        """
+        bin_s = '{:032b}'.format(int(s, 16))  # Convert string to 32-bit binary string
+        major_version = int(bin_s[4:10], 2)
+        minor_version = int(bin_s[10:16], 2)
+        build_number = int(bin_s[17:32], 2)
+        return cls(major_version=major_version, minor_version=minor_version, major_build=build_number)
+
     def api_version(self):
         if EXCHANGE_2013_SP1 <= self < EXCHANGE_2016:
             return 'Exchange2013_SP1'
@@ -114,6 +125,9 @@ class Build(PickleMixIn):
         except KeyError:
             raise ValueError('API version for build %s is unknown' % self)
 
+    def fullname(self):
+        return VERSIONS[self.api_version()][1]
+
     def __cmp__(self, other):
         # __cmp__ is not a magic method in Python3. We'll just use it here to implement comparison operators
         c = (self.major_version > other.major_version) - (self.major_version < other.major_version)
@@ -169,12 +183,10 @@ EXCHANGE_2019 = Build(15, 2)
 EXCHANGE_O365 = Build(15, 20)
 
 
-@python_2_unicode_compatible
-class Version(PickleMixIn):
-    """
-    Holds information about the server version
-    """
-    __slots__ = ('build', 'api_version')
+class Version:
+    """Holds information about the server version."""
+
+    __slots__ = 'build', 'api_version'
 
     def __init__(self, build, api_version=None):
         if not isinstance(build, (Build, type(None))):
@@ -183,7 +195,7 @@ class Version(PickleMixIn):
         if api_version is None:
             self.api_version = build.api_version()
         else:
-            if not isinstance(api_version, string_types):
+            if not isinstance(api_version, str):
                 raise ValueError("'api_version' must be a string")
             self.api_version = api_version
 
@@ -192,20 +204,24 @@ class Version(PickleMixIn):
         return VERSIONS[self.api_version][1]
 
     @classmethod
-    def guess(cls, protocol, hint=None):
-        """
-        Tries to ask the server which version it has. We haven't set up an Account object yet, so we generate requests
+    def guess(cls, protocol, api_version_hint=None):
+        """Ask the server which version it has. We haven't set up an Account object yet, so we generate requests
         by hand. We only need a response header containing a ServerVersionInfo element.
 
         To get API version and build numbers from the server, we need to send a valid SOAP request. We can't do that
         without a valid API version. To solve this chicken-and-egg problem, we try all possible API versions that this
         package supports, until we get a valid response.
+
+        :param protocol:
+        :param api_version_hint:  (Default value = None)
         """
         from .services import ResolveNames
         # The protocol doesn't have a version yet, so default to latest supported version if we don't have a hint.
-        api_version = hint or API_VERSIONS[0]
+        api_version = api_version_hint or API_VERSIONS[0]
         log.debug('Asking server for version info using API version %s', api_version)
-        # We don't know the build version yet. Hopefully, the server will report it in the SOAP header
+        # We don't know the build version yet. Hopefully, the server will report it in the SOAP header. Lots of
+        # places expect a version to have a build, so this is a bit dangerous, but passing a fake build around is also
+        # dangerous. Make sure the call to ResolveNames does not require a version build.
         protocol.config.version = Version(build=None, api_version=api_version)
         # Use ResolveNames as a minimal request to the server to test if the version is correct. If not, ResolveNames
         # will try to guess the version automatically.
@@ -217,6 +233,8 @@ class Version(PickleMixIn):
         except ResponseMessageError:
             # We survived long enough to get a new version
             pass
+        if not protocol.version.build:
+            raise AttributeError('Protocol should have a build number at this point')
         return protocol.version
 
     @staticmethod
@@ -239,14 +257,13 @@ class Version(PickleMixIn):
             if cls._is_invalid_version_string(api_version_from_server):
                 # For unknown reasons, Office 365 may respond with an API version strings that is invalid in a request.
                 # Detect these so we can fallback to a valid version string.
-                log.info('API version "%s" worked but server reports version "%s". Using "%s"', requested_api_version,
-                         api_version_from_server, requested_api_version)
+                log.debug('API version "%s" worked but server reports version "%s". Using "%s"', requested_api_version,
+                          api_version_from_server, requested_api_version)
                 api_version_from_server = requested_api_version
             else:
-                # Work around a bug in Exchange that reports a bogus API version in the XML response. Trust server
-                # response except 'V2_nn' or 'V201[5,6]_nn_mm' which is bogus
-                log.info('API version "%s" worked but server reports version "%s". Using "%s"', requested_api_version,
-                         api_version_from_server, api_version_from_server)
+                # Trust API version from server response
+                log.debug('API version "%s" worked but server reports version "%s". Using "%s"', requested_api_version,
+                          api_version_from_server, api_version_from_server)
         return cls(build=build, api_version=api_version_from_server)
 
     def __eq__(self, other):
